from dataclasses import dataclass


@dataclass(frozen=True)
class ResourceMetadata:
    """A resource type is identified by a triple of (server, service, type)."""

    server: str
    service: str
    type: str

    def __str__(self) -> str:
        return f"{self.server}://{self.service}/{self.type}"


@dataclass(frozen=True)
class PromptMetadata:
    """A prompt is identified by server, service, and type."""

    server: str
    service: str
    type: str

    def __str__(self) -> str:
        return f"{self.server}_{self.service}_{self.type}"


@dataclass(frozen=True)
class ToolMetadata:
    """A tool is identified by server, service, and name."""

    server: str
    service: str
    name: str

    def __str__(self) -> str:
        return f"{self.server}_{self.service}_{self.name}"



"""Utilities around Pydantic settings."""

import json
import logging
from pathlib import Path
from typing import TypeVar

from pydantic import ValidationError
from pydantic_settings import BaseSettings
from rich.panel import Panel
from rich.text import Text

from starbridge.utils.console import console

T = TypeVar("T", bound=BaseSettings)

logger = logging.getLogger(__name__)


def load_settings(settings_class: type[T]) -> T:
    """Load settings with error handling and nice formatting.

    Args:
        settings_class: The Pydantic settings class to instantiate

    Returns:
        (T): Instance of the settings class

    Raises:
        SystemExit: If settings validation fails
    """
    try:
        return settings_class()
    except ValidationError as e:
        errors = json.loads(e.json())
        text = Text()
        text.append(
            "Validation error(s): \n\n",
            style="debug",
        )

        prefix = settings_class.model_config.get("env_prefix", "")
        for error in errors:
            env_var = f"{prefix}{error['loc'][0]}".upper()
            logger.fatal(f"Configuration invalid! {env_var}: {error['msg']}")
            text.append(f"• {env_var}", style="yellow bold")
            text.append(f": {error['msg']}\n")

        text.append(
            "\nCheck settings defined in the process environment and in file ",
            style="info",
        )
        env_file = str(settings_class.model_config.get("env_file", ".env") or ".env")
        text.append(
            str(Path(__file__).parent.parent.parent.parent / env_file),
            style="bold blue underline",
        )

        console.print(
            Panel(
                text,
                title="Configuration invalid!",
                border_style="error",
            )
        )
        exit(78)



"""Define styling for typer, overriding defaults"""

from rich.console import Console
from rich.theme import Theme

console = Console(
    theme=Theme({
        "logging.level.info": "purple4",
        "debug": "light_cyan3",
        "info": "purple4",
        "warning": "yellow1",
        "error": "red1",
    })
)



from .console import console
from .logging import get_logger
from .settings import load_settings
from .signature import description_and_params

__all__ = ["console", "get_logger", "description_and_params", "load_settings"]



import pytest
from typer.testing import CliRunner

from starbridge.cli import cli


@pytest.fixture
def runner():
    return CliRunner()


def test_built_with_love(runner):
    """Check epilog shown."""
    result = runner.invoke(cli, ["--help"])
    assert result.exit_code == 0
    assert (
        "Built with love in Berlin by Helmut Hoffer von Ankershoffen" in result.stdout
    )


def test_invalid_command(runner):
    """Test invalid command returns error"""
    result = runner.invoke(cli, ["invalid"])
    assert result.exit_code != 0



from collections import defaultdict
from dataclasses import dataclass
from importlib.metadata import entry_points
from inspect import signature
from typing import Any
from urllib.parse import urlparse

import mcp.types as types
import typer

from starbridge.mcp.context import MCPContext
from starbridge.mcp.models import ResourceMetadata
from starbridge.utils.signature import description_and_params


@dataclass(frozen=True)
class ResourceType:
    """A resource type is identified by a triple of (server, service, type)"""

    server: str
    service: str
    type: str

    def __str__(self) -> str:
        return f"{self.server}://{self.service}/{self.type}"


class MCPBaseService:
    """Base class for MCP services."""

    @property
    def service_prefix(self) -> str:
        """
        Automatically generate service prefix from the module path.
        Example: 'starbridge.confluence.service.Service' becomes 'starbridge-confluence-'
        """
        # Get full module path of the actual service class
        module_path = self.__class__.__module__
        # Take the first three parts (e.g., 'starbridge.confluence.service')
        # and convert dots to dashes
        prefix = "_".join(module_path.split(".")[:2]) + "_"
        return prefix

    @staticmethod
    def get_services() -> list[type["MCPBaseService"]]:
        """Discover all registered MCP services."""
        services = []
        for entry_point in entry_points(group="starbridge.services"):
            try:
                service_class = entry_point.load()
                if issubclass(service_class, MCPBaseService):
                    services.append(service_class)
            except Exception as e:
                print(f"Error loading service {entry_point.name}: {e}")
        return services

    @staticmethod
    def get_cli() -> tuple[str | None, typer.Typer | None]:
        """Get CLI name and typer for this service if available.
        Returns a tuple of (name, typer) or (None, None) if no CLI available."""
        return None, None

    def info(self) -> dict:
        """Get info about configuration of this service. Override in subclass."""
        raise NotImplementedError

    def health(self, context: MCPContext | None = None) -> str:
        """Get health of this service. Override in subclass."""
        raise NotImplementedError

    def tool_list(self, context: MCPContext | None = None) -> list[types.Tool]:
        """Get available tools. Discovers tools by looking for methods decorated with @mcp_tool."""
        tools = []
        for method_name in dir(self.__class__):
            method = getattr(self.__class__, method_name)
            if hasattr(method, "__mcp_tool__"):
                meta = method.__mcp_tool__
                description, required, params = description_and_params(method)
                tools.append(
                    types.Tool(
                        name=str(meta),  # Use metadata string representation
                        description=description,
                        inputSchema={
                            "type": "object",
                            "required": required,
                            "properties": params,
                        },
                    )
                )
        return tools

    def resource_list(self, context: MCPContext | None = None) -> list[types.Resource]:
        """Get available resources by discovering and calling all resource iterators."""
        resources = []
        type_map = defaultdict(list)  # renamed from prefix_map

        # Find all resource iterators
        for method_name in dir(self.__class__):
            method = getattr(self.__class__, method_name)
            if hasattr(method, "__mcp_resource_iterator__"):
                meta = method.__mcp_resource_iterator__
                if not meta.type:  # renamed from prefix
                    raise ValueError(
                        f"Resource iterator {method_name} missing required type"
                    )

                # Check type uniqueness
                type_map[meta.type].append(method_name)
                if len(type_map[meta.type]) > 1:
                    raise ValueError(
                        f"Multiple resource iterators found for type '{meta.type}': {type_map[meta.type]}"
                    )

                # Get resources from iterator
                iterator_resources = method(self, context)

                # Validate each resource URI
                for resource in iterator_resources:
                    parsed = urlparse(str(resource.uri))
                    if parsed.scheme != meta.server:
                        raise ValueError(
                            f"Resource URI scheme '{parsed.scheme}' doesn't match decorator scheme '{meta.server}'"
                        )
                    if parsed.netloc != meta.service:
                        raise ValueError(
                            f"Resource URI service '{parsed.netloc}' doesn't match decorator service '{meta.service}'"
                        )
                    if not parsed.path.startswith(
                        f"/{meta.type}/"
                    ):  # renamed from prefix
                        raise ValueError(
                            f"Resource URI path doesn't start with '/{meta.type}/'"
                        )

                resources.extend(iterator_resources)

        return resources

    def resource_type_list(
        self, context: MCPContext | None = None
    ) -> set[ResourceMetadata]:
        """Get available resource types by discovering all resource iterators."""
        types = set()
        for method_name in dir(self.__class__):
            method = getattr(self.__class__, method_name)
            if hasattr(method, "__mcp_resource_iterator__"):
                meta = method.__mcp_resource_iterator__
                if meta.type:
                    types.add(meta)
        return types

    # Remove resource_get method as it's now handled by MCPServer

    def prompt_list(self, context: MCPContext | None = None) -> list[types.Prompt]:
        """Get available prompts by discovering decorated prompt methods."""
        prompts = []
        for method_name in dir(self.__class__):
            method = getattr(self.__class__, method_name)
            if hasattr(method, "__mcp_prompt__"):
                meta = method.__mcp_prompt__
                sig = signature(method)
                description, required, params = description_and_params(method)

                # Convert signature params to PromptArguments
                arguments = []
                for name, _param in sig.parameters.items():
                    if name in ("self", "context"):
                        continue
                    arguments.append(
                        types.PromptArgument(
                            name=name,
                            description=params.get(name, {}).get(
                                "description", f"Parameter {name}"
                            ),
                            required=name in required,
                        )
                    )

                prompts.append(
                    types.Prompt(
                        name=str(meta),
                        description=description,
                        arguments=arguments,
                    )
                )
        return prompts

    def get_prompt(
        self,
        name: str,
        arguments: dict[str, Any] | None,
        context: MCPContext | None = None,
    ) -> types.GetPromptResult:
        """Get prompt content. Override in subclass."""
        return types.GetPromptResult(description=None, messages=[])



from .cli import cli, serve
from .context import MCPContext
from .decorators import mcp_prompt, mcp_resource, mcp_resource_iterator, mcp_tool
from .models import PromptMetadata, ResourceMetadata
from .server import MCPServer
from .service import MCPBaseService

__all__ = [
    "serve",
    "cli",
    "MCPServer",
    "MCPContext",
    "MCPBaseService",
    "mcp_tool",
    "mcp_resource",
    "mcp_resource_iterator",
    "mcp_prompt",
    "ResourceMetadata",
    "PromptMetadata",
]



"""
CLI to interact with Claude Desktop application
"""

import pathlib
import subprocess
from typing import Annotated

import typer

from starbridge.base import __project_name__
from starbridge.utils.console import console

from .service import Service

cli = typer.Typer(no_args_is_help=True)


@cli.command()
def health():
    """Health of Claude"""
    console.print(Service().health())


@cli.command()
def info():
    """Info about Claude"""
    console.print(Service().info())


@cli.command()
def config():
    """Print config of Claude Desktop application"""
    if not Service.is_installed():
        console.print(
            f"Claude Desktop application is not installed at '{Service.application_directory()}' - you can install it from https://claude.ai/download"
        )
        return
    if not Service.config_path().is_file():
        console.print(f"No config file found at '{Service.config_path()}'")
        return
    console.print(f"Printing config file at '{Service.config_path()}'")
    console.print_json(data=Service.config_read())


@cli.command()
def log(
    tail: Annotated[
        bool,
        typer.Option(
            help="Tail logs",
        ),
    ] = False,
    last: Annotated[
        int,
        typer.Option(help="Number of lines to show"),
    ] = 100,
    name: Annotated[
        str,
        typer.Option(
            help="Name of the MCP server - use 'main' for main mcp.log of Claude Desktop application",
        ),
    ] = __project_name__,
):
    """Show logs."""
    log_path = Service.log_path(name if name != "main" else None)
    size = pathlib.Path(log_path).stat().st_size
    human_size = (
        f"{size / 1024 / 1024:.1f}MB" if size > 1024 * 1024 else f"{size / 1024:.1f}KB"
    )
    console.print(
        f"Showing max {last} lines of log at '{log_path}' ({human_size}{', tailing' if tail else ''})",
    )
    if tail:
        subprocess.run(
            [
                "tail",
                "-n",
                str(last),
                "-f",
                Service.log_path(name if name != "main" else None),
            ],
            check=False,
        )
    else:
        subprocess.run(
            [
                "tail",
                "-n",
                str(last),
                Service.log_path(name if name != "main" else None),
            ],
            check=False,
        )


@cli.command(name="restart")
def restart():
    """Restart Claude Desktop application"""
    if not Service.is_installed():
        console.print(
            f"Claude Desktop application is not installed at '{Service.application_directory()}' - you can install it from https://claude.ai/download"
        )
        return
    Service().restart()
    console.print("Claude Desktop application was restarted")



import logfire
from pydantic_settings import BaseSettings, SettingsConfigDict

from starbridge.base import __project_name__, __version__
from starbridge.instrumentation.otel_mcp_instrumentation import MCPInstrumentor
from starbridge.utils.settings import load_settings


class LogfireSettings(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix=f"{__project_name__.upper()}_LOGFIRE_",
        extra="ignore",
        env_file=".env",
        env_file_encoding="utf-8",
    )

    token: str | None = None
    environment: str = "default"
    instrument_mcp_enabled: bool = False


def logfire_initialize():
    settings = load_settings(LogfireSettings)

    if settings.token is None:
        return False

    def scrubbing_callback(m: logfire.ScrubMatch):
        if (
            m.path == ("attributes", "rpc.jsonrpc.result")
            and m.pattern_match.group(0) == "session"
        ):
            return m.value

        if (
            m.path == ("attributes", "jsonrpc.result")
            and m.pattern_match.group(0) == "session"
        ):
            return m.value

    logfire.configure(
        send_to_logfire="if-token-present",
        token=settings.token,
        environment=settings.environment,
        service_name=__project_name__,
        console=False,
        code_source=logfire.CodeSource(
            repository="https://github.com/helmut-hoffer-von-ankershoffen/starbridge",
            revision=__version__,
            root_path="",
        ),
        scrubbing=logfire.ScrubbingOptions(callback=scrubbing_callback),
    )
    logfire.instrument_system_metrics(base="full")

    logfire.install_auto_tracing(
        modules=["starbridge.confluence"], min_duration=0.001
    )  # FIXME: get modules from settings

    if settings.instrument_mcp_enabled:
        MCPInstrumentor().instrument()

    return True



from typing import Any, Literal

from mcp.shared.context import RequestContext
from pydantic import AnyUrl, BaseModel


class MCPContext(BaseModel):
    """Context object providing access to MCP capabilities.

    This provides a cleaner interface to MCP's RequestContext functionality.
    It gets injected into tool and resource functions that request it via type hints.

    To use context in a tool function, add a parameter with the Context type annotation:

    ```python
    @server.tool()
    def my_tool(
        x: int,
        ctx: Context,
    ) -> str:
        # Log messages to the client
        ctx.info(
            f"Processing {x}"
        )
        ctx.debug(
            "Debug info"
        )
        ctx.warning(
            "Warning message"
        )
        ctx.error(
            "Error message"
        )

        # Report progress
        ctx.report_progress(
            50, 100
        )

        # Access resources
        data = ctx.read_resource(
            "resource://data"
        )

        # Get request info
        request_id = ctx.request_id
        client_id = ctx.client_id

        return str(x)
    ```

    The context parameter name can be anything as long as it's annotated with Context.
    The context is optional - tools that don't need it can omit the parameter.
    """

    _request_context: RequestContext | None
    _mcp: Any | None  # Avoid circular import by using Any

    def __init__(
        self,
        *,
        request_context: RequestContext | None = None,
        mcp: Any | None = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self._request_context = request_context
        self._mcp = mcp

    @property
    def mcp(self) -> Any:
        """Access to the MCP server."""
        if self._mcp is None:
            raise ValueError("Context is not available outside of a request")
        return self._mcp

    @property
    def request_context(self) -> RequestContext:
        """Access to the underlying request context."""
        if self._request_context is None:
            raise ValueError("Context is not available outside of a request")
        return self._request_context

    async def report_progress(
        self, progress: float, total: float | None = None
    ) -> None:
        """Report progress for the current operation.

        Args:
            progress: Current progress value e.g. 24
            total: Optional total value e.g. 100
        """

        progress_token = (
            self.request_context.meta.progressToken
            if self.request_context.meta
            else None
        )

        if not progress_token:
            return

        await self.request_context.session.send_progress_notification(
            progress_token=progress_token, progress=progress, total=total
        )

    async def read_resource(self, uri: str | AnyUrl) -> str | bytes:
        """Read a resource by URI.

        Args:
            uri: Resource URI to read

        Returns:
            (str | bytes): The resource content as either text or bytes
        """
        assert self._mcp is not None, "Context is not available outside of a request"
        return await self._mcp.read_resource(uri)

    async def log(
        self,
        level: Literal["debug", "info", "warning", "error"],
        message: str,
        *,
        logger_name: str | None = None,
    ) -> None:
        """Send a log message to the client.

        Args:
            level: Log level (debug, info, warning, error)
            message: Log message
            logger_name: Optional logger name
            **extra: Additional structured data to include
        """
        await self.request_context.session.send_log_message(
            level=level, data=message, logger=logger_name
        )

    @property
    def client_id(self) -> str | None:
        """Get the client ID if available."""
        return (
            getattr(self.request_context.meta, "client_id", None)
            if self.request_context.meta
            else None
        )

    @property
    def request_id(self) -> str:
        """Get the unique ID for this request."""
        return str(self.request_context.request_id)

    @property
    def session(self):
        """Access to the underlying session for advanced usage."""
        return self.request_context.session

    # Convenience methods for common log levels
    async def debug(self, message: str, **extra: Any) -> None:
        """Send a debug log message."""
        await self.log("debug", message, **extra)

    async def info(self, message: str, **extra: Any) -> None:
        """Send an info log message."""
        await self.log("info", message, **extra)

    async def warning(self, message: str, **extra: Any) -> None:
        """Send a warning log message."""
        await self.log("warning", message, **extra)

    async def error(self, message: str, **extra: Any) -> None:
        """Send an error log message."""
        await self.log("error", message, **extra)



from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from functools import wraps

import mcp.server.stdio
from mcp import JSONRPCError, JSONRPCRequest, JSONRPCResponse
from mcp.types import JSONRPCNotification
from opentelemetry import trace
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.trace import StatusCode

tracer = trace.get_tracer("mcp.server")


class MCPInstrumentor(BaseInstrumentor):
    def instrumentation_dependencies(self):
        return []

    def _instrument(self, **kwargs):
        self._transaction_spans = {}
        original_stdio_server = mcp.server.stdio.stdio_server

        @asynccontextmanager
        @wraps(original_stdio_server)
        async def instrumented_stdio_server(*args, **kwargs):
            async with original_stdio_server(*args, **kwargs) as (
                read_stream,
                write_stream,
            ):
                traced_read = TracedReceiveStream(
                    read_stream, tracer, self._transaction_spans
                )
                traced_write = TracedSendStream(
                    write_stream, tracer, self._transaction_spans
                )
                yield traced_read, traced_write

        mcp.server.stdio.stdio_server = instrumented_stdio_server

    def _uninstrument(self, **kwargs):
        pass


def _handle_transaction(tracer, msg, span_kind, active_spans):
    """Handle span lifecycle for a JSON-RPC request/response transaction"""
    root = msg.root
    msg_id = getattr(root, "id", None)

    if isinstance(root, JSONRPCRequest):
        # Start new transaction span
        span = tracer.start_span(
            f"mcp.{span_kind.lower()}.transaction.{root.method}",
            kind=getattr(trace.SpanKind, span_kind.upper()),
        )
        _set_request_attributes(span, msg)
        active_spans[msg_id] = span
    elif isinstance(root, JSONRPCResponse | JSONRPCError):
        # End existing transaction span
        span = active_spans.pop(msg_id, None)
        if span:
            _set_response_attributes(span, msg)
            span.end()


def _handle_notification(tracer, msg, span_kind):
    """Handle span for a JSON-RPC notification"""
    with tracer.start_span(
        f"mcp.{span_kind.lower()}.notification.{msg.root.method}",
        kind=getattr(trace.SpanKind, span_kind.upper()),
    ) as span:
        if isinstance(msg.root, JSONRPCNotification):
            span.set_attribute("jsonrpc.notification.method", msg.root.method)
            if hasattr(msg.root, "params"):
                span.set_attribute("jsonrpc.notification.params", str(msg.root.params))
            span.set_status(StatusCode.OK)


def _set_request_attributes(span, msg):
    """Set span attributes for a JSON-RPC request"""
    root = msg.root
    if isinstance(root, JSONRPCRequest):
        span.set_attribute("jsonrpc.request.id", root.id)
        span.set_attribute("jsonrpc.request.method", root.method)
        if hasattr(root, "params"):
            span.set_attribute("jsonrpc.request.params", str(root.params))


def _set_response_attributes(span, msg):
    """Set span attributes for a JSON-RPC response"""
    root = msg.root
    if isinstance(root, JSONRPCResponse):
        span.set_attribute("jsonrpc.response.result", str(root.result))
        span.set_status(StatusCode.OK)
    elif isinstance(root, JSONRPCError):
        span.set_attribute("jsonrpc.error.code", root.error.code)
        span.set_attribute("jsonrpc.error.message", root.error.message)
        span.set_status(StatusCode.ERROR, root.error.message)


class TracedSendStream:
    def __init__(self, stream, tracer, active_spans):
        self._stream = stream
        self._tracer = tracer
        self._active_spans = active_spans

    async def __aenter__(self):
        await self._stream.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self._stream.__aexit__(exc_type, exc_val, exc_tb)

    async def send(self, msg):
        root = getattr(msg, "root", None)

        if isinstance(root, JSONRPCNotification):
            _handle_notification(
                self._tracer, msg, "CLIENT"
            )  # We're sending a notification
        else:
            _handle_transaction(
                self._tracer, msg, "CLIENT", self._active_spans
            )  # We're sending a request/response

        await self._stream.send(msg)

    def __getattr__(self, attr):
        return getattr(self._stream, attr)


class TracedReceiveStream:
    def __init__(self, stream, tracer, active_spans):
        self._stream = stream
        self._tracer = tracer
        self._active_spans = active_spans
        self._current_span = None

    async def __aenter__(self):
        await self._stream.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self._stream.__aexit__(exc_type, exc_val, exc_tb)

    async def receive(self):
        msg = await self._stream.receive()
        root = getattr(msg, "root", None)

        if isinstance(root, JSONRPCNotification):
            _handle_notification(
                self._tracer, msg, "SERVER"
            )  # It's an incoming notification
        else:
            _handle_transaction(
                self._tracer, msg, "SERVER", self._active_spans
            )  # It's an incoming request/response

        return msg

    def __aiter__(self) -> AsyncIterator:
        return TracedAsyncIterator(
            self._stream.__aiter__(), self._tracer, self._active_spans
        )

    def __getattr__(self, attr):
        return getattr(self._stream, attr)


class TracedAsyncIterator:
    def __init__(self, iterator: AsyncIterator, tracer, active_spans):
        self._iterator = iterator
        self._tracer = tracer
        self._active_spans = active_spans

    def __aiter__(self):
        return self

    async def __anext__(self):
        msg = await self._iterator.__anext__()
        root = getattr(msg, "root", None)

        if isinstance(root, JSONRPCNotification):
            _handle_notification(self._tracer, msg, "CLIENT")
        else:
            _handle_transaction(self._tracer, msg, "SERVER", self._active_spans)

        return msg
